(* large cardinals, solyga@gmx.de, 2022-09-14 *)
IMPLEMENTATION MODULE LC;

FROM SYSTEM IMPORT BYTE;
IMPORT Types, Lib, Lib2;
IMPORT IO;

TYPE
  TCard = Types.TCard;
  TInt = Types.TInt;


PROCEDURE WrHex( x-: ARRAY OF BYTE );
VAR
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF CardToHex( x, s ) THEN
    IO.WrStr( s );
  ELSE
    IO.WrStr( '???' );
  END;
END WrHex;

CONST
  WrLn ::= IO.WrLn;
  WrStr ::= IO.WrStr;
  WrCard ::= IO.WrCard;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Lib.FatalError( s );
  HALT;
END Fatal;


PROCEDURE StrLen( s-: ARRAY OF CHAR ): CARDINAL;
VAR i: CARDINAL;
BEGIN
  FOR i := 0 TO SIZE( s ) - 1 DO
    IF s[i] = 0C THEN RETURN i END;
  END;
  RETURN SIZE( s );
END StrLen;

PROCEDURE Invert( VAR s: ARRAY OF CHAR; l: CARDINAL );
VAR
  i, j: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( s ) THEN l := SIZE( s ) END;
  IF l < 2 THEN RETURN END;
  FOR i := 0 TO l DIV 2 - 1 DO
    j := l - 1 - i;
    c := s[ i ]; s[ i ] := s[ j ]; s[ j ] := c;
  END;
END Invert;

PROCEDURE Char( x: TSht ): CHAR;
BEGIN
  IF x < 10 THEN
    RETURN CHR( x + ORD('0') );
  ELSIF x < bMax THEN
    RETURN CHR( x - 10 + ORD('A') );
  ELSE
    RETURN '?';  (* ??? *)
  END;
END Char;

PROCEDURE Digit( c: CHAR ): TSht;
BEGIN
  c := CAP( c );
  CASE c OF
  | '0'..'9': RETURN ORD( c ) - ORD( '0' );
  | 'A'..'Z': RETURN ORD( c ) - ORD( 'A' ) + 10;
  ELSE
    RETURN bMax;
  END;
END Digit;


PROCEDURE CardToHexOld( x-: ARRAY OF BYTE; VAR s: ARRAY OF CHAR ): BOOLEAN;
VAR
  l: CARDINAL;
  ok: BOOLEAN;

  PROCEDURE Put( c: CHAR );
  BEGIN
    IF l >= SIZE( s ) THEN ok := FALSE; RETURN END;
    s[ l ] := c;
    INC( l );
  END Put;

VAR
  i: CARDINAL;
  y: TSht;
BEGIN
  l := 0;
  ok := TRUE;
  FOR i := 0 TO SIZE( x ) - 1 DO
    y := x[ i ];
    Put( Char( y MOD 10H ) );
    Put( Char( y DIV 10H ) );
  END;
  WHILE ( l > 1 ) & ( s[l-1] = '0' ) DO DEC( l ) END;
  IF l < SIZE( s ) THEN s[ l ] := 0C END;
  Invert( s, l );
  RETURN ok;
END CardToHexOld;

PROCEDURE CardToHex( x-: ARRAY OF BYTE; VAR s: ARRAY OF CHAR ): BOOLEAN;
VAR
  ok: BOOLEAN;
  l: CARDINAL;

  PROCEDURE Put( y: TSht );
  BEGIN
    IF ( y = 0 ) & ( l = 0 ) THEN RETURN END;
    IF l >= SIZE( s ) THEN ok := FALSE; RETURN END;
    s[ l ] := Char( y );
    INC( l );
  END Put;

VAR
  i: CARDINAL;
  y: TSht;
BEGIN
  ok := TRUE;
  l := 0;
  i := SIZE( x );
  WHILE i > 0 DO
    DEC( i );
    y := x[ i ];
    Put( y DIV 10H );
    Put( y MOD 10H );
  END;
  IF l = 0 THEN s[ 0 ] := '0'; l := 1 END;
  IF l < SIZE( s ) THEN s[ l ] := 0C END;
  RETURN ok;
END CardToHex;


PROCEDURE HexToCard( s-: ARRAY OF CHAR; VAR x: ARRAY OF BYTE ): BOOLEAN;
VAR
  ok: BOOLEAN;
  l: CARDINAL;

  PROCEDURE Get( VAR x: TSht );
  BEGIN
    IF l > 0 THEN
      DEC( l );
      x := Digit( s[l] );
      IF x >= 10H THEN x := 0; ok := FALSE END;
    ELSE
      x := 0;
    END;
  END Get;

VAR
  i: CARDINAL;
  y, z: TSht;
BEGIN
  ok := TRUE;
  l := StrLen( s );
  FOR i := 0 TO SIZE( x ) - 1 DO
    Get( y );
    Get( z );
    x[ i ] := BYTE( y + z * 10H );
  END;
  RETURN ok;
END HexToCard;


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Len( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 1 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len;

PROCEDURE Len0( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 0 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len0;

PROCEDURE LenI( x-: ARRAY OF BYTE; i: CARDINAL ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  IF l <= i THEN RETURN 0 END;
  WHILE ( l > i ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l - i;
END LenI;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Cpy( x-: ARRAY OF BYTE; VAR y: ARRAY OF BYTE ): BOOLEAN;
VAR
  l, i: CARDINAL;
BEGIN
  l := Len( x );
  IF l > SIZE( y ) THEN RETURN FALSE END;
  FOR i := 0 TO l - 1 DO y[i] := x[i] END;
  WHILE l < SIZE( y ) DO y[l] := BYTE(0); INC( l ) END;
  RETURN TRUE;
END Cpy;

PROCEDURE Cmp( x-, y-: ARRAY OF BYTE ): Types.TShtInt;
VAR
  lx, ly: CARDINAL;
  cx, cy: TSht;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx < ly THEN RETURN -1 ELSIF lx > ly THEN RETURN 1 END;
  WHILE lx > 0 DO
    DEC( lx );
    cx := x[lx]; cy := y[lx];
    IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
  END;
  RETURN 0;
END Cmp;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Inc( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TCard;
    | 1: l, h: TSht;
    END;
  END;
VAR
  i, l: CARDINAL;
  c: TCnvDesc;
BEGIN
  i := 0; l := Len( y ); c.h := 0;
  WHILE i < l DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TSht( x[i] ) + TSht( y[i] ) + c.h;
    x[i] := c.l;
    INC( i );
  END;
  WHILE c.h > 0 DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TSht( x[i] ) + c.h;
    x[i] := c.l;
    INC( i );
  END;
  RETURN TRUE;
END Inc;

PROCEDURE Add( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Cpy( x, z ) & Inc( z, y );
END Add;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TInt;
    | 1: l, h: TSht;
    END;
  END;
VAR
  i, l: CARDINAL;
  c: TCnvDesc;
BEGIN
  i := 0; l := Len0( y ); c.h := 0;
  WHILE i < l DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TInt( TSht( x[i] ) ) - TSht( y[i] ) - c.h;
    IF c.v < 0 THEN c.h := 1 END;
    x[i] := c.l;
    INC( i );
  END;
  WHILE c.h > 0 DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TInt( TSht( x[i] ) ) - c.h;
    IF c.v < 0 THEN c.h := 1 END;
    x[i] := c.l;
    INC( i );
  END;
  RETURN TRUE;
END Dec;

PROCEDURE Sub( x, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Dec( x, y ) & Cpy( x, z );
END Sub;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Mul( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TCard;
    | 1: l, h: TSht;
    END;
  END;
VAR
  ix, lx, iy, ly, i: CARDINAL;
  f: TSht;
  c: TCnvDesc;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx + ly - 1 > SIZE( z ) THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  FOR iy := 0 TO ly - 1 DO
    f := TSht( y[iy] );
    c.h := 0;
    FOR ix := 0 TO lx - 1 DO
      i := ix + iy;
      c.v := TSht( z[i] ) + TSht( x[ix] ) * f + c.h;  (* FF + FF * FF + FF = FFFF *)
      z[i] := c.l;
    END;
    IF c.h > 0 THEN
      i := lx + iy;
      IF i >= SIZE( z ) THEN RETURN FALSE END;
      z[i] := c.h;
    END;
  END;
  RETURN TRUE;
END Mul;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Div( x, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Cmp( x-: ARRAY OF BYTE; ix: CARDINAL; y-: ARRAY OF BYTE ): INTEGER;
  VAR
    lx, ly: CARDINAL;
    cx, cy: TSht;
  BEGIN
    lx := LenI( x, ix ); ly := Len0( y );
    IF lx < ly THEN RETURN -1 ELSIF lx > ly THEN RETURN 1 END;
    WHILE lx > 0 DO
      DEC( lx );
      cx := x[ix+lx]; cy := y[lx];
      IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
    END;
    RETURN 0;
  END Cmp;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; ix: CARDINAL; y-: ARRAY OF BYTE );
  TYPE
    TCnvDesc = RECORD
      CASE : CARDINAL OF
      | 0: v: TInt;
      | 1: l, h: TSht;
      END;
    END;
  VAR
    iy, ly: CARDINAL;
    c: TCnvDesc;
  BEGIN
    iy := 0; ly := Len0( y ); c.h := 0;
    WHILE iy < ly DO
      c.v := TInt( TSht( x[ix] ) ) - TSht( y[iy] ) - c.h;
      IF c.v < 0 THEN c.h := 1 END;
      x[ix] := c.l;
      INC( ix ); INC( iy );
    END;
    IF c.h > 0 THEN x[ix] := BYTE( TSht( x[ix] ) - 1 ) END;
  END Dec;

  PROCEDURE Inc( VAR x: BYTE );
  BEGIN
    x := BYTE( TSht(x) + 1 );
  END Inc;

VAR
  l: CARDINAL;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  l := Len( x );
  WHILE l > 0 DO
    DEC( l );
    WHILE Cmp( x, l, y ) >= 0 DO
      Dec( x, l, y );
      Inc( z[l] );
    END;
  END;
  RETURN Cpy( x, r );
END Div;
<* POP *>


PROCEDURE Mpl( x, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Mul( x, y, z );
END Mpl;

PROCEDURE StrToCard( s-: ARRAY OF CHAR; VAR x: ARRAY OF BYTE; b: TSht ): BOOLEAN;

  PROCEDURE Eval( c: CHAR; VAR x: TSht ): BOOLEAN;
  BEGIN
    x := Digit( c );
    RETURN x < b;
  END Eval;

VAR
  i, l: CARDINAL;
  y: TSht;
BEGIN
  IF ( b < bMin ) OR ( b > bMax ) THEN RETURN FALSE END;
  l := StrLen( s );
  IF l = 0 THEN RETURN FALSE END;
  Lib2.Fill( x, 0 );
  FOR i := 0 TO l - 1 DO
    IF ~ Eval( s[i], y ) OR ~ Mpl( x, b, x ) OR ~ Inc( x, y ) THEN RETURN FALSE END;
  END;
  RETURN TRUE;
END StrToCard;


PROCEDURE Dvd( x, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Div( x, y, z, r );
END Dvd;

PROCEDURE CardToStr( x: ARRAY OF BYTE; VAR s: ARRAY OF CHAR; b: TSht ): BOOLEAN;
VAR
  l: CARDINAL;

  PROCEDURE Put( c: CHAR ): BOOLEAN;
  BEGIN
    IF l >= SIZE( s ) THEN RETURN FALSE END;
    s[ l ] := c;
    INC( l );
    RETURN TRUE;
  END Put;

VAR
  r: TSht;
BEGIN
  IF ( b < bMin ) OR ( b > bMax ) THEN RETURN FALSE END;
  l := 0;
  REPEAT
    IF ~ Dvd( x, b, x, r ) OR ~ Put( Char(r) ) THEN RETURN FALSE END;
  UNTIL Len0( x ) = 0;
  IF l < SIZE( s ) THEN s[ l ] := 0C END;
  Invert( s, l );
  RETURN TRUE;
END CardToStr;

END LC.
